@@ -32,6 +32,7 @@
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <limits.h>
 #include <pwd.h>
 #include <stdlib.h>
 #include <string.h>
@@ -48,6 +49,7 @@
 #include "pathnames.h"
 
 #define TMAX(a,b) ((a)>(b)?(a):(b))
+#define TMIN(a,b) ((a)<(b)?(a):(b))
 
 /* size of the event structure, not counting name */
 #define EVENT_SIZE  (sizeof (struct inotify_event))
@@ -152,10 +154,41 @@ check_orphans(cron_db *db) {
 	}
 }
 
+static int
+find_orphan(const char *uname, const char *fname, const char *tabname) {
+	orphan *o;
+
+	for (o = orphans; o != NULL; o = o->next) {
+		if (uname && o->uname) {
+			if (strcmp(uname, o->uname) != 0)
+				continue;
+		} else if (uname != o->uname)
+			continue;
+
+		if (fname && o->fname) {
+			if (strcmp(fname, o->fname) != 0)
+				continue;
+		} else if (fname != o->fname)
+			continue;
+
+		if (tabname && o->tabname) {
+			if (strcmp(tabname, o->tabname) != 0)
+				continue;
+		} else if (tabname != o->tabname)
+			continue;
+		return 1;
+	}
+
+	return 0;
+}
+
 static void
 add_orphan(const char *uname, const char *fname, const char *tabname) {
 	orphan *o;
 
+	if (find_orphan(uname, fname, tabname))
+		return;
+
 	o = calloc(1, sizeof(*o));
 	if (o == NULL)
 		return;
@@ -206,6 +239,8 @@ process_crontab(const char *uname, const char *fname, const char *tabname,
 	if ((crontab_fd = check_open(tabname, uname, pw, &mtime)) == -1)
 		goto next_crontab;
 
+	mtime = TMIN(new_db->mtime, mtime);
+
 	Debug(DLOAD, ("\t%s:", fname));
 
 	if (old_db != NULL)
@@ -230,7 +265,7 @@ process_crontab(const char *uname, const char *fname, const char *tabname,
 		 * we finish with the crontab...
 		 */
 		Debug(DLOAD, (" [delete old data]"));
-			unlink_user(old_db, u);
+		unlink_user(old_db, u);
 		free_user(u);
 		log_it(fname, getpid(), "RELOAD", tabname, 0);
 	}
@@ -252,7 +287,7 @@ process_crontab(const char *uname, const char *fname, const char *tabname,
 static int
 cluster_host_is_local(void)
 {
-	char filename[MAXNAMLEN+1];
+	char filename[NAME_MAX+1];
 	int is_local;
 	FILE *f;
 	char hostname[MAXHOSTNAMELEN], myhostname[MAXHOSTNAMELEN];
@@ -297,18 +332,18 @@ void check_inotify_database(cron_db * old_db) {
 	cron_db new_db;
 	DIR_T *dp;
 	DIR *dir;
-	struct timeval time;
+	struct timeval tv;
 	fd_set rfds;
 	int retval;
 	char buf[BUF_LEN];
 	pid_t pid = getpid();
-	time.tv_sec = 0;
-	time.tv_usec = 0;
+	tv.tv_sec = 0;
+	tv.tv_usec = 0;
 
 	FD_ZERO(&rfds);
 	FD_SET(old_db->ifd, &rfds);
 
-	retval = select(old_db->ifd + 1, &rfds, NULL, NULL, &time);
+	retval = select(old_db->ifd + 1, &rfds, NULL, NULL, &tv);
 	if (retval == -1) {
 		if (errno != EINTR)
 			log_it("CRON", pid, "INOTIFY", "select failed", errno);
@@ -317,6 +352,7 @@ void check_inotify_database(cron_db * old_db) {
 	else if (FD_ISSET(old_db->ifd, &rfds)) {
 		new_db.head = new_db.tail = NULL;
 		new_db.ifd = old_db->ifd;
+		new_db.mtime = time(NULL) - 1;
 		while ((retval = read(old_db->ifd, buf, sizeof (buf))) == -1 &&
 			errno == EINTR) ;
 
@@ -348,7 +384,7 @@ void check_inotify_database(cron_db * old_db) {
 		}
 		else {
 			while (NULL != (dp = readdir(dir))) {
-				char tabname[MAXNAMLEN + 1];
+				char tabname[NAME_MAX + 1];
 
 				if (not_a_crontab(dp))
 					continue;
@@ -366,12 +402,12 @@ void check_inotify_database(cron_db * old_db) {
 		}
 		else {
 			while (NULL != (dp = readdir(dir))) {
-				char fname[MAXNAMLEN + 1], tabname[MAXNAMLEN + 1];
+				char fname[NAME_MAX + 1], tabname[NAME_MAX + 1];
 
 				if (not_a_crontab(dp))
 					continue;
 
-				strncpy(fname, dp->d_name, MAXNAMLEN);
+				strncpy(fname, dp->d_name, NAME_MAX);
 
 				if (!glue_strings(tabname, sizeof tabname, SPOOL_DIR,
 						dp->d_name, '/'))
@@ -421,14 +457,17 @@ int load_database(cron_db * old_db) {
 	DIR *dir;
 	pid_t pid = getpid();
 	int is_local = 0;
+	time_t now;
 
 	Debug(DLOAD, ("[%ld] load_database()\n", (long) pid));
 
-		/* before we start loading any data, do a stat on SPOOL_DIR
-		 * so that if anything changes as of this moment (i.e., before we've
-		 * cached any of the database), we'll see the changes next time.
-		 */
-		if (stat(SPOOL_DIR, &statbuf) < OK) {
+	now = time(NULL);
+
+	/* before we start loading any data, do a stat on SPOOL_DIR
+	 * so that if anything changes as of this moment (i.e., before we've
+	 * cached any of the database), we'll see the changes next time.
+	 */
+	if (stat(SPOOL_DIR, &statbuf) < OK) {
 		log_it("CRON", pid, "STAT FAILED", SPOOL_DIR, errno);
 		statbuf.st_mtime = 0;
 	}
@@ -461,13 +500,17 @@ int load_database(cron_db * old_db) {
 	 * Note that old_db->mtime is initialized to 0 in main(), and
 	 * so is guaranteed to be different than the stat() mtime the first
 	 * time this function is called.
+	 *
+	 * We also use now - 1 as the upper bound of timestamp to avoid race,
+	 * when a crontab is updated twice in a single second when we are
+         * just reading it.
 	 */
-	if (old_db->mtime == TMAX(crond_stat.st_mtime,
-			TMAX(statbuf.st_mtime, syscron_stat.st_mtime))
+	if (old_db->mtime == TMIN(now - 1, TMAX(crond_stat.st_mtime,
+			TMAX(statbuf.st_mtime, syscron_stat.st_mtime)))
 		) {
 		Debug(DLOAD, ("[%ld] spool dir mtime unch, no load needed.\n",
 				(long) pid));
-			return 0;
+		return 0;
 	}
 
 	/* something's different.  make a new database, moving unchanged
@@ -475,8 +518,7 @@ int load_database(cron_db * old_db) {
 	 * actually changed.  Whatever is left in the old database when
 	 * we're done is chaff -- crontabs that disappeared.
 	 */
-	new_db.mtime = TMAX(crond_stat.st_mtime,
-		TMAX(statbuf.st_mtime, syscron_stat.st_mtime));
+	new_db.mtime = now - 1;
 	new_db.head = new_db.tail = NULL;
 #if defined WITH_INOTIFY
 	new_db.ifd = old_db->ifd;
@@ -492,7 +534,7 @@ int load_database(cron_db * old_db) {
 	}
 	else {
 		while (NULL != (dp = readdir(dir))) {
-			char tabname[MAXNAMLEN + 1];
+			char tabname[NAME_MAX + 1];
 
 			if (not_a_crontab(dp))
 				continue;
@@ -519,12 +561,12 @@ int load_database(cron_db * old_db) {
 		is_local = cluster_host_is_local();
 
 		while (is_local && NULL != (dp = readdir(dir))) {
-			char fname[MAXNAMLEN + 1], tabname[MAXNAMLEN + 1];
+			char fname[NAME_MAX + 1], tabname[NAME_MAX + 1];
 
 			if (not_a_crontab(dp))
 				continue;
 
-			strncpy(fname, dp->d_name, MAXNAMLEN);
+			strncpy(fname, dp->d_name, NAME_MAX);
 
 			if (!glue_strings(tabname, sizeof tabname, SPOOL_DIR, fname, '/'))
 				continue;	/* XXX log? */
@@ -601,7 +643,7 @@ static int not_a_crontab(DIR_T * dp) {
 
 	len = strlen(dp->d_name);
 
-	if (len >= MAXNAMLEN)
+	if (len >= NAME_MAX)
 		return (1);	/* XXX log? */
 
 	if ((len > 0) && (dp->d_name[len - 1] == '~'))
@@ -628,7 +670,7 @@ static void max_mtime(const char *dir_name, struct stat *max_st) {
 	}
 
 	while (NULL != (dp = readdir(dir))) {
-		char tabname[MAXNAMLEN + 1];
+		char tabname[NAME_MAX + 1];
 
 		if ( not_a_crontab ( dp ) && strcmp(dp->d_name, CRON_HOSTNAME) != 0)
 			continue;
